/* Transforms a table of lines "III 0xUUUU COMMENTS" where III is an index
   to a table of lines "0xAABB 0xUUUU" where AA, BB, ... are byte values.
   Usage: index2table type < whatwg-table.txt > TABLE.TXT
 */
#include <stdio.h>
#include <stdlib.h>

static void eightbit_table (void)
{
  /* One-dimensional type. III = 0..127 maps to 0x80..0xFF. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  for (;;) {
    int c = getc (stdin);
    if (c == EOF)
      break;
    if (c == '\n')
      continue;
    if (c != '#') {
      int a, b;
      ungetc (c, stdin);
      if (fscanf (stdin, "%d 0x%x", &a, &b) != 2) {
        fprintf (stderr, "Unexpected syntax at %ld\n", ftell (stdin));
        exit (1);
      }
      printf ("0x%02X\t0x%04X\n", 128 + a, b);
    }
    do c = getc (stdin); while (c != EOF && c != '\n');
  }
}

static void euc_jp_table (const char *jisx0208_filename, const char *jisx0212_filename)
{
  /* Two-dimensional type for EUC-JP.
     Read two tables: the one for JISX0208 and the one for JISX0212. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  /* See https://encoding.spec.whatwg.org/#euc-jp-decoder */
  {
    int i;
    for (i = 0xa1; i < 0xe0; i++)
      printf ("0x%02X%02X\t0x%04X\n", 0x8e, i, 0xff61+(i-0xa1));
  }
  { /* III = 0..8835 maps to 0x{A1..FE}{A1..FE}. */
    const char *filename = jisx0208_filename;
    FILE *fp = fopen (filename, "r");
    for (;;) {
      int c = getc (fp);
      if (c == EOF)
        break;
      if (c == '\n')
        continue;
      if (c != '#') {
        int a, b, a1, a2;
        ungetc (c, fp);
        if (fscanf (fp, "%d 0x%x", &a, &b) != 2) {
          fprintf (stderr, "Unexpected syntax at %s:%ld\n", filename, ftell (fp));
          exit (1);
        }
        a1 = a / 94;
        a2 = a % 94;
        if (a1 < 94) {
          printf ("0x%02X%02X\t0x%04X\n", 0xa1 + a1, 0xa1 + a2, b);
        }
      }
      do c = getc (fp); while (c != EOF && c != '\n');
    }
    fclose (fp);
  }
  { /* III = 0..8835 maps to 0x8F{A1..FE}{A1..FE}. */
    const char *filename = jisx0212_filename;
    FILE *fp = fopen (filename, "r");
    for (;;) {
      int c = getc (fp);
      if (c == EOF)
        break;
      if (c == '\n')
        continue;
      if (c != '#') {
        int a, b, a1, a2;
        ungetc (c, fp);
        if (fscanf (fp, "%d 0x%x", &a, &b) != 2) {
          fprintf (stderr, "Unexpected syntax at %s:%ld\n", filename, ftell (fp));
          exit (1);
        }
        a1 = a / 94;
        a2 = a % 94;
        if (a1 < 94) {
          printf ("0x%02X%02X%02X\t0x%04X\n", 0x8f, 0xa1 + a1, 0xa1 + a2, b);
        }
      }
      do c = getc (fp); while (c != EOF && c != '\n');
    }
    fclose (fp);
  }
}

static void shift_jis_table (void)
{
  /* Two-dimensional type for Shift_JIS.
     III = 0..11103 maps to 0x{81..9F,E0..F9}{40..7E,80..FC}. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  /* See https://encoding.spec.whatwg.org/#shift_jis-decoder */
  {
    int i;
    for (i = 0xa1; i < 0xe0; i++)
      printf ("0x%02X\t0x%04X\n", i, 0xff61+(i-0xa1));
  }
  {
    int a;
    for (a = 8836; a < 10716; a++) {
      int a1 = a / 188, a2 = a % 188, b = 0xe000+(a-8836);
      printf ("0x%02X%02X\t0x%04X\n", (a1 < 31 ? 0x81 : 0xc1) + a1, (a2 < 63 ? 0x40 : 0x41) + a2, b);
    }
  }
  for (;;) {
    int c = getc (stdin);
    if (c == EOF)
      break;
    if (c == '\n')
      continue;
    if (c != '#') {
      int a, b, a1, a2;
      ungetc (c, stdin);
      if (fscanf (stdin, "%d 0x%x", &a, &b) != 2) {
        fprintf (stderr, "Unexpected syntax at %ld\n", ftell (stdin));
        exit (1);
      }
      a1 = a / 188;
      a2 = a % 188;
      printf ("0x%02X%02X\t0x%04X\n", (a1 < 31 ? 0x81 : 0xc1) + a1, (a2 < 63 ? 0x40 : 0x41) + a2, b);
    }
    do c = getc (stdin); while (c != EOF && c != '\n');
  }
}

static void gbk_table (void)
{
  /* Two-dimensional type for CP936.
     III = 0..23939 maps to 0x{81..FE}{40..7E,80..FE}. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  /* See https://encoding.spec.whatwg.org/#gbk */
  printf ("0x%02X\t0x%04X\n", 0x80, 0x20ac);
  for (;;) {
    int c = getc (stdin);
    if (c == EOF)
      break;
    if (c == '\n')
      continue;
    if (c != '#') {
      int a, b, a1, a2;
      ungetc (c, stdin);
      if (fscanf (stdin, "%d 0x%x", &a, &b) != 2) {
        fprintf (stderr, "Unexpected syntax at %ld\n", ftell (stdin));
        exit (1);
      }
      a1 = a / 190;
      a2 = a % 190;
      printf ("0x%02X%02X\t0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x41) + a2, b);
    }
    do c = getc (stdin); while (c != EOF && c != '\n');
  }
}

static void gb18030_table (const char *gbk_table, const char *gb18030_ranges_table)
{
  /* Two-dimensional type for GB18030.
     III = 0..23939 maps to 0x{81..FE}{40..7E,80..FE}. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  /* See https://encoding.spec.whatwg.org/#gb18030 */
  {
    const char *filename = gbk_table;
    FILE *fp = fopen (filename, "r");
    for (;;) {
      int c = getc (fp);
      if (c == EOF)
        break;
      if (c == '\n')
        continue;
      if (c != '#') {
        int a, b, a1, a2;
        ungetc (c, fp);
        if (fscanf (fp, "%d 0x%x", &a, &b) != 2) {
          fprintf (stderr, "Unexpected syntax at %s:%ld\n", filename, ftell (fp));
          exit (1);
        }
        a1 = a / 190;
        a2 = a % 190;
        printf ("0x%02X%02X\t0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x41) + a2, b);
      }
      do c = getc (fp); while (c != EOF && c != '\n');
    }
    fclose (fp);
  }
  {
    const char *filename = gb18030_ranges_table;
    FILE *fp = fopen (filename, "r");
    int a = 0, b = 0;
    for (;;) {
      int c = getc (fp);
      if (c == EOF)
        break;
      if (c == '\n')
        continue;
      if (c == '#') {
        do c = getc (fp); while (c != EOF && c != '\n');
      } else {
        int next_a, next_b;
        ungetc (c, fp);
        if (fscanf (fp, "%d 0x%x\n", &next_a, &next_b) != 2) {
          fprintf (stderr, "Unexpected syntax at %s:%ld\n", filename, ftell (fp));
          exit (1);
        }
        while (a < next_a && b < next_b) {
          int a1 = a / 12600;
          int a2 = (a / 1260) % 10;
          int a3 = (a / 10) % 126;
          int a4 = a % 10;
          printf ("0x%02X%02X%02X%02X\t0x%04X\n", 0x81 + a1, 0x30 + a2, 0x81 + a3, 0x30 + a4, b);
          a++, b++;
        }
        a = next_a; b = next_b;
      }
    }
    while (b < 0x110000) {
      int a1 = a / 12600;
      int a2 = (a / 1260) % 10;
      int a3 = (a / 10) % 126;
      int a4 = a % 10;
      printf ("0x%02X%02X%02X%02X\t0x%04X\n", 0x81 + a1, 0x30 + a2, 0x81 + a3, 0x30 + a4, b);
      a++, b++;
    }
    fclose (fp);
  }
}

static void big5_table (void)
{
  /* Two-dimensional type for CP950.
     III = 0..19781 maps to 0x{81..FE}{40..7E,A1..FE}. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  /* See https://encoding.spec.whatwg.org/#big5-decoder */
  {
    int a = 1133, b1 = 0x00ca, b2 = 0x0304;
    int a1 = a / 157, a2 = a % 157;
    printf ("0x%02X%02X\t0x%04X 0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x62) + a2, b1, b2);
  }
  {
    int a = 1135, b1 = 0x00ca, b2 = 0x030c;
    int a1 = a / 157, a2 = a % 157;
    printf ("0x%02X%02X\t0x%04X 0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x62) + a2, b1, b2);
  }
  {
    int a = 1164, b1 = 0x00ea, b2 = 0x0304;
    int a1 = a / 157, a2 = a % 157;
    printf ("0x%02X%02X\t0x%04X 0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x62) + a2, b1, b2);
  }
  {
    int a = 1166, b1 = 0x00ea, b2 = 0x030c;
    int a1 = a / 157, a2 = a % 157;
    printf ("0x%02X%02X\t0x%04X 0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x62) + a2, b1, b2);
  }
  for (;;) {
    int c = getc (stdin);
    if (c == EOF)
      break;
    if (c == '\n')
      continue;
    if (c != '#') {
      int a, b, a1, a2;
      ungetc (c, stdin);
      if (fscanf (stdin, "%d 0x%x", &a, &b) != 2) {
        fprintf (stderr, "Unexpected syntax at %ld\n", ftell (stdin));
        exit (1);
      }
      a1 = a / 157;
      a2 = a % 157;
      printf ("0x%02X%02X\t0x%04X\n", 0x81 + a1, (a2 < 63 ? 0x40 : 0x62) + a2, b);
    }
    do c = getc (stdin); while (c != EOF && c != '\n');
  }
}

static void euc_kr_table (void)
{
  /* Two-dimensional type for CP949.
     III = 0..23749 maps to 0x{81..FD}{41..FE}. */
  {
    int i;
    for (i = 0; i < 0x80; i++)
      printf ("0x%02X\t0x%04X\n", i, i);
  }
  for (;;) {
    int c = getc (stdin);
    if (c == EOF)
      break;
    if (c == '\n')
      continue;
    if (c != '#') {
      int a, b, a1, a2;
      ungetc (c, stdin);
      if (fscanf (stdin, "%d 0x%x", &a, &b) != 2) {
        fprintf (stderr, "Unexpected syntax at %ld\n", ftell (stdin));
        exit (1);
      }
      a1 = a / 190;
      a2 = a % 190;
      printf ("0x%02X%02X\t0x%04X\n", 0x81 + a1, 0x41 + a2, b);
    }
    do c = getc (stdin); while (c != EOF && c != '\n');
  }
}

int main (int argc, char *argv[])
{
  int type;

  if (argc < 2) {
    fprintf (stderr, "Usage: %s type\n", argv[0]);
    exit (1);
  }
  type = atoi (argv[1]);

  if (type == 1) {
    eightbit_table ();
  } else if (type == 2) {
    euc_jp_table (argv[2], argv[3]);
  } else if (type == 3) {
    shift_jis_table ();
  } else if (type == 4) {
    gbk_table ();
  } else if (type == 5) {
    gb18030_table (argv[2], argv[3]);
  } else if (type == 6) {
    big5_table ();
  } else if (type == 7) {
    euc_kr_table ();
  } else {
    fprintf (stderr, "Unrecognized type %d\n", type);
    exit (1);
  }

  return 0;
}
